"""
    Cursor

Module for controlling the console cursor using console escape codes.
"""
module Cursor

using ..ConsoleCodes; const CC = ConsoleCodes


"""
    Cursor.up(y)

Move the console cursor up by `y`.

Behavior is undefined for `y ≤ 0` but no check is performed.
See [`Cursor.translate`](@ref) for a safe version of this operation.
"""
up(io::IO, y::Integer) = CC.control(io, 'A', y)
up(y::Integer) = up(stdout, y)

"""
    Cursor.down(y)

Move the console cursor down by `y`.

Behavior is undefined for `y ≤ 0` but no check is performed.
See [`Cursor.translate`](@ref) for a safe version of this operation.
"""
down(io::IO, y::Integer) = CC.control(io, 'B', y)
down(y::Integer) = down(stdout, y)

"""
    Cursor.right(x)

Move the cnosle cursor right by `x`.

Behavior is undefined for `y ≤ 0` but no check is performed.
See [`Cursor.translate`](@ref) for a safe version of this operation.
"""
right(io::IO, x::Integer) = CC.control(io, 'C', x)
right(x::Integer) = right(stdout, x)

"""
    Cursor.left(x)

Move the console cursor left by `x`.

Behavior is undefined for `y ≤ 0` but no check is performed.
See [`Cursor.translate`](@ref) for a safe version of this operation.
"""
left(io::IO, x::Integer) = CC.control(io, 'D', x)
left(x::Integer) = left(stdout, x)

"""
    Cursor.down1(y)

Move the console cursor down by `y`, to the start of the line.

Invalid for `y ≤ 0`.
"""
function down1(io::IO, y::Integer)
    if y < 1
        throw(ArgumentError("down1 requires positive definite argument, got $y"))
    end
    CC.control(io, 'E', y)
end
down1(y::Integer) = down1(stdout, y)

"""
    Cursor.up1(y)

Move the consle cursor up by `y`, to the start of the line.

Invalid for `y ≤ 0`.
"""
function up1(io::IO, y::Integer)
    if y < 1
        throw(ArgumentError("up1 requires positive definite argument, got $y"))
    end
    CC.control(io, 'F', y)
end
up1(y::Integer) = up1(stdout, y)

"""
    Cursor.xtranslate(x)

Translate the cursor horizontally by `x`.
"""
function xtranslate(io::IO, x::Integer)
    if x > 0
        right(io, x)
    elseif x < 0
        left(io, -x)
    else
    end
end
xtranslate(x::Integer) = xtranslate(stdout, x)

"""
    Cursor.ytranslate(y)

Translate the cursor vertically by `y`.
"""
function ytranslate(io::IO, y::Integer)
    if y > 0
        up(io, y)
    elseif y < 0
        down(io, -y)
    else
    end
end
ytranslate(y::Integer) = ytranslate(stdout, y)

"""
    Cursor.translate(x, y)

Translate the cursor by `x` horizontally and `y` vertically.
"""
function translate(io::IO, x::Integer, y::Integer)
    xtranslate(io, x)
    ytranslate(io, y)
end
translate(x::Integer, y::Integer) = translate(stdout, x, y)

"""
    Cursor.tocolumn(x)

Move the console cursor to column `x`.
"""
tocolumn(io::IO, x::Integer) = CC.control(io, 'G', y)
tocolumn(x::Integer) = tocolumn(stdout, x)

"""
    Cursor.to(x, y)

Move the console cursor to position `(x, y)`.
"""
to(io::IO, x::Integer, y::Integer) = CC.control(io, 'H', x, y)
to(x::Integer, y::Integer) = to(stdout, x, y)

"""
    Cursor.save()

Save the cursor position.  The state is saved to an internal OS register that can be
loaded with [`Cursor.restore`](@ref) but is otherwise inaccessible.
"""
save(io::IO) = CC.control(io, 's')
save() = save(stdout)

"""
    Cursor.restore()

Restore the cursor position saved with [`Cursor.save()`](@ref).
"""
restore(io::IO) = CC.control(io, 'u')
restore() = restore(stdout)


end
